/*++
/* NAME
/*	dict_cidr 3
/* SUMMARY
/*	Dictionary interface for CIDR data
/* SYNOPSIS
/*	#include <dict_cidr.h>
/*
/*	DICT	*dict_cidr_open(name, open_flags, dict_flags)
/*	const char *name;
/*	int	open_flags;
/*	int	dict_flags;
/* DESCRIPTION
/*	dict_cidr_open() opens the named file and stores
/*	the key/value pairs where the key must be either a
/*	"naked" IP address or a netblock in CIDR notation.
/* SEE ALSO
/*	dict(3) generic dictionary manager
/* AUTHOR(S)
/*	Jozsef Kadlecsik
/*	kadlec@blackhole.kfki.hu
/*	KFKI Research Institute for Particle and Nuclear Physics
/*	POB. 49
/*	1525 Budapest, Hungary
/*
/*	Wietse Venema
/*	IBM T.J. Watson Research
/*	P.O. Box 704
/*	Yorktown Heights, NY 10598, USA
/*
/*	Dean C. Strik
/*	Department ICT Services
/*	Eindhoven University of Technology
/*	P.O. Box 513
/*	5600 MB  Eindhoven, Netherlands
/*	E-mail: <dean@ipnet6.org>
/*--*/

/* System library. */

#include <sys_defs.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef INET6
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#endif

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif

/* Utility library. */

#include <mymalloc.h>
#include <msg.h>
#include <vstream.h>
#include <vstring.h>
#include <stringops.h>
#include <readlline.h>
#include <dict.h>
#include <dict_cidr.h>
#include <match_ops.h>

/* Application-specific. */

 /*
  * Each rule in a CIDR table is parsed and stored in a linked list.
  */
typedef struct DICT_CIDR_ENTRY {
    ADDR_PATTERN *pattern;		/* address pattern structure */
    char   *value;			/* lookup result */
    struct DICT_CIDR_ENTRY *next;	/* next entry */
} DICT_CIDR_ENTRY;

typedef struct {
    DICT    dict;			/* generic members */
    DICT_CIDR_ENTRY *head;		/* first entry */
} DICT_CIDR;

#define BITS_PER_ADDR_V4   32
#define BITS_PER_ADDR_V6   128

/* dict_cidr_lookup - CIDR table lookup */

static const char *dict_cidr_lookup(DICT *dict, const char *key)
{
    char   *myname = "dict_cidr_lookup";

    DICT_CIDR *dict_cidr = (DICT_CIDR *) dict;
    DICT_CIDR_ENTRY *entry;
#ifdef INET6
    struct addrinfo hints, *res0;
    int     aierr;
#else
    struct sockaddr_in sin;
#endif

    if (msg_verbose)
	msg_info("%s: %s: %s", myname, dict_cidr->dict.name, key);

#ifdef INET6
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = PF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_NUMERICHOST;
    /*
     * Since access maps call the CIDR map first with the
     * hostname and only then with the addresses, we just
     * return 0 when an entry isn't numeric, as expressed
     * by the EAI_NONAME error.
     */
    aierr = getaddrinfo(key, NULL, &hints, &res0);
    if (aierr == EAI_NONAME) {
	if (msg_verbose)
	    msg_info("%s: non-address key \"%s\"",
		     myname, key);
	return (0);
    }
    if (aierr != 0)
	msg_fatal("%s: getaddrinfo(%s): %s",
		  myname, key, GAI_STRERROR(aierr));
    for (entry = dict_cidr->head; entry; entry = entry->next)
	if (match_sockaddr(res0->ai_addr,
			   entry->pattern->addr,
			   entry->pattern->masklen)) {
	    freeaddrinfo(res0);
	    return (entry->value);
	}
    freeaddrinfo(res0);
    return (0);

#else /* INET6 */

    memset(&sin, 0, sizeof(sin));
    sin.family = AF_INET;
    sin.addr = inet_addr(key);
    if (sin.addr == INADDR_NONE)
	return (0);
    for (entry = dict_cidr->head; entry; entry = entry->next)
	if (match_sockaddr(&sa, entry->pattern->addr, entry->pattern->masklen))
	    return (entry->value);
    return (0);

#endif
}

/* dict_cidr_close - close the CIDR table */

static void dict_cidr_close(DICT *dict)
{
    DICT_CIDR *dict_cidr = (DICT_CIDR *) dict;
    DICT_CIDR_ENTRY *entry;
    DICT_CIDR_ENTRY *next;

    for (entry = dict_cidr->head; entry; entry = next) {
	next = entry->next;
	addr_pattern_free(entry->pattern);
	myfree(entry->value);
	myfree((char *) entry);
    }
    dict_free(dict);
}

/* dict_cidr_parse_rule - parse CIDR table rule into network, mask and value */

static DICT_CIDR_ENTRY *dict_cidr_parse_rule(const char *mapname, int lineno,
					             char *p)
{
    DICT_CIDR_ENTRY *rule;
    char   *key;
    char   *value;
    ADDR_PATTERN *pattern;
    VSTRING *lookup_err;
    int    lookup_res;

    /*
     * Split the rule into key and value. We already eliminated leading
     * whitespace, comments, empty lines or lines with whitespace only. This
     * means a null key can't happen but we will handle this anyway.
     */
    key = p;
    while (*p && !ISSPACE(*p))			/* Skip over key */
	p++;
    if (*p)					/* Terminate key */
	*p++ = 0;
    while (*p && ISSPACE(*p))			/* Skip whitespace */
	p++;
    value = p;
    trimblanks(value, 0)[0] = 0;		/* Trim trailing blanks */
    if (*key == 0) {
	msg_warn("cidr map %s, line %d: no address pattern: skipping this rule",
		 mapname, lineno);
	return (0);
    }
    if (*value == 0) {
	msg_warn("cidr map %s, line %d: no lookup result: skipping this rule",
		 mapname, lineno);
	return (0);
    }

    /*
     * We rewrite the key to standard notation, and check the validity of
     * the pattern.
     * We cannot use MATCH_FLAG_STRICT_ADDR since access checks try not only
     * the numerical address but the resolved hostname as well.
     */
    lookup_err = vstring_alloc(100);
    lookup_res = std_addr_pattern(MATCH_FLAG_NOLOOKUP |
				  MATCH_FLAG_NONNULL_HOST,
				  key, &pattern, lookup_err);
    if (pattern == NULL) {
	if (lookup_res == 0 && VSTRING_LEN(lookup_err) != 0)
	    msg_warn("cidr map %s, line %d: %s: skipping this rule",
		     mapname, lineno, vstring_str(lookup_err));
	vstring_free(lookup_err);
	return (0);
    }
    vstring_free(lookup_err);

    /*
     * Bundle up the result.
     */
    rule = (DICT_CIDR_ENTRY *) mymalloc(sizeof(DICT_CIDR_ENTRY));
    rule->pattern = pattern;
    rule->value = mystrdup(value);
    rule->next = 0;

    if (msg_verbose)
	msg_info("dict_cidr_open: %s: %s/%d %s",
		 mapname, pattern->pattern, pattern->masklen, rule->value);

    return (rule);
}

/* dict_cidr_open - parse CIDR table */

DICT   *dict_cidr_open(const char *mapname, int open_flags, int dict_flags)
{
    DICT_CIDR *dict_cidr;
    VSTREAM *map_fp;
    VSTRING *line_buffer = vstring_alloc(100);
    DICT_CIDR_ENTRY *rule;
    DICT_CIDR_ENTRY *last_rule = 0;
    int     lineno = 0;

    /*
     * Sanity checks.
     */
    if (open_flags != O_RDONLY)
	msg_fatal("%s:%s map requires O_RDONLY access mode",
		  DICT_TYPE_CIDR, mapname);

    /*
     * XXX Eliminate unnecessary queries by setting a flag that says "this
     * map matches network addresses only".
     */
    dict_cidr = (DICT_CIDR *) dict_alloc(DICT_TYPE_CIDR, mapname,
					 sizeof(*dict_cidr));
    dict_cidr->dict.lookup = dict_cidr_lookup;
    dict_cidr->dict.close = dict_cidr_close;
    dict_cidr->dict.flags = dict_flags | DICT_FLAG_PATTERN;
    dict_cidr->head = 0;

    if ((map_fp = vstream_fopen(mapname, O_RDONLY, 0)) == 0)
	msg_fatal("open %s: %m", mapname);

    while (readlline(line_buffer, map_fp, &lineno)) {
	rule = dict_cidr_parse_rule(mapname, lineno, vstring_str(line_buffer));
	if (rule == 0)
	    continue;
	if (last_rule == 0)
	    dict_cidr->head = rule;
	else
	    last_rule->next = rule;
	last_rule = rule;
    }

    /*
     * Clean up.
     */
    if (vstream_fclose(map_fp))
	msg_fatal("cidr map %s: read error: %m", mapname);
    vstring_free(line_buffer);

    return (DICT_DEBUG (&dict_cidr->dict));
}
